__version__ = "0.7.0"


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

from pandabear.decorators import check, check_schemas, dataframe_check
from pandabear.model import DataFrameModel, SeriesModel
from pandabear.model_components import Field, Index
from pandabear.typehints import DataFrame, Series
